@@ -1,6 +1,6 @@
 /* vim: set sw=8: -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * gsf-output.c: interface for used by the ole layer to read raw data
+ * gsf-output.c: interface for storing data
  *
  * Copyright (C) 2002 Jody Goldberg (jody@gnome.org)
  *
@@ -25,6 +25,9 @@
 
 #define GET_CLASS(instance) G_TYPE_INSTANCE_GET_CLASS (instance, GSF_OUTPUT_TYPE, GsfOutputClass)
 
+static gboolean
+gsf_output_vprintf (GsfOutput *output, char const* format, va_list args);
+
 static void
 gsf_output_finalize (GObject *obj)
 {
@@ -38,14 +41,16 @@ gsf_output_finalize (GObject *obj)
 	if (!output->is_closed)
 		g_warning ("unrefing an unclosed stream");
 
-	if (output->name != NULL) {
-		g_free (output->name);
-		output->name = NULL;
-	}
+	g_free (output->name);
+	output->name = NULL;
+	g_free (output->printf_buf);
+	output->printf_buf = NULL;
+
 	if (output->container != NULL) {
 		g_object_unref (G_OBJECT (output->container));
 		output->container = NULL;
 	}
+	g_clear_error (&output->err);
 }
 
 static void
@@ -56,15 +61,21 @@ gsf_output_init (GObject *obj)
 	output->cur_offset	= 0;
 	output->cur_size	= 0;
 	output->name		= NULL;
-	output->container	= NULL;
 	output->wrapped_by	= NULL;
+	output->container	= NULL;
+	output->err		= NULL;
 	output->is_closed	= FALSE;
+	output->printf_buf	= NULL;
+	output->printf_buf_size = 0;
 }
 
 static void
 gsf_output_class_init (GObjectClass *gobject_class)
 {
+	GsfOutputClass  *output_class  = GSF_OUTPUT_CLASS (gobject_class);
+
 	gobject_class->finalize = gsf_output_finalize;
+	output_class->Vprintf   = gsf_output_vprintf;
 }
 
 GSF_CLASS_ABSTRACT (GsfOutput, gsf_output,
@@ -116,19 +127,24 @@ gsf_output_size (GsfOutput *output)
  * @output :
  *
  * Close a stream.
- * Returns TRUE if things are successful.
+ * Returns FALSE on error
  **/
 gboolean
 gsf_output_close (GsfOutput *output)
 {
-	g_return_val_if_fail (output != NULL, FALSE);
-	g_return_val_if_fail (!output->is_closed, FALSE);
+	gboolean res;
 
-	if (GET_CLASS (output)->Close (output)) {
-		output->is_closed = TRUE;
-		return TRUE;
-	}
-	return FALSE;
+	g_return_val_if_fail (GSF_IS_OUTPUT (output),
+		gsf_output_set_error (output, 0, "<internal>"));
+	g_return_val_if_fail (!output->is_closed,
+		gsf_output_set_error (output, 0, "<internal>"));
+
+	/* The implementation will log any errors, but we can never try to
+	 * close multiple times even on failure.
+	 */
+	res = GET_CLASS (output)->Close (output);
+	output->is_closed = TRUE;
+	return res;
 }
 
 /**
@@ -164,41 +180,46 @@ gsf_output_tell	(GsfOutput *output)
  * @offset :
  * @whence :
  *
- * Returns TRUE on error.
+ * Returns FALSE on error.
  **/
 gboolean
 gsf_output_seek	(GsfOutput *output, gsf_off_t offset, GSeekType whence)
 {
 	gsf_off_t pos = offset;
 
-	g_return_val_if_fail (output != NULL, -1);
+	g_return_val_if_fail (output != NULL, FALSE);
 
 	switch (whence) {
 	case G_SEEK_SET: break;
 	case G_SEEK_CUR: pos += output->cur_offset;	break;
 	case G_SEEK_END: pos += output->cur_size;	break;
-	default : return TRUE;
+	default :
+		g_warning ("Invalid seek type %d", (int)whence);
+		return FALSE;
 	}
 
-	if (pos < 0)
-		return TRUE;
+	if (pos < 0) {
+		g_warning ("Invalid seek position %" GSF_OFF_T_FORMAT
+			   ", which is before the start of the file", pos);
+		return FALSE;
+	}
 
 	/* If we go nowhere, just return.  This in particular handles null
 	 * seeks for streams with no seek method.
 	 */
 	if (pos == output->cur_offset)
-		return FALSE;
-
-	if (GET_CLASS (output)->Seek (output, offset, whence))
 		return TRUE;
 
-	output->cur_offset = pos;
-
-	/* NOTE : it is possible for the current pos to be beyond the end of
-	 * the file.  The intervening space is not filled with 0 until
-	 * something is written.
-	 */
+	if (GET_CLASS (output)->Seek (output, offset, whence)) {
+		/* NOTE : it is possible for the current pos to be beyond the
+		 * end of the file.  The intervening space is not filled with 0
+		 * until something is written.
+		 */
+		output->cur_offset = pos;
+		return TRUE;
+	}
 
+	/* the implementation should have assigned whatever errors are necessary */
 	return FALSE;
 }
 
@@ -224,15 +245,31 @@ gsf_output_write (GsfOutput *output,
 			output->cur_size = output->cur_offset;
 		return TRUE;
 	}
+
+	/* the implementation should have assigned whatever errors are necessary */
 	return FALSE;
 }
 
+/**
+ * gsf_output_error :
+ * @output :
+ *
+ * Returns the last error logged on the output, or NULL.
+ **/
+GError const *
+gsf_output_error (GsfOutput const *output)
+{
+	g_return_val_if_fail (GSF_IS_OUTPUT (output), NULL);
+	return output->err;
+}
+
 /**
  * gsf_output_set_name :
  * @output :
  * @name :
  *
- * protected.
+ * <protected> This is a utility routine that should only be used by derived
+ * outputs.
  *
  * Returns : TRUE if the assignment was ok.
  **/
@@ -241,7 +278,7 @@ gsf_output_set_name (GsfOutput *output, char const *name)
 {
 	char *buf;
 
-	g_return_val_if_fail (output != NULL, FALSE);
+	g_return_val_if_fail (GSF_IS_OUTPUT (output), FALSE);
 
 	buf = g_strdup (name);
 	if (output->name != NULL)
@@ -255,12 +292,15 @@ gsf_output_set_name (GsfOutput *output, char const *name)
  * @output :
  * @container :
  *
+ * <protected> This is a utility routine that should only be used by derived
+ * outputs.
+ *
  * Returns : TRUE if the assignment was ok.
  */
 gboolean
 gsf_output_set_container (GsfOutput *output, GsfOutfile *container)
 {
-	g_return_val_if_fail (output != NULL, FALSE);
+	g_return_val_if_fail (GSF_IS_OUTPUT (output), FALSE);
 
 	if (container != NULL)
 		g_object_ref (G_OBJECT (container));
@@ -270,6 +310,39 @@ gsf_output_set_container (GsfOutput *output, GsfOutfile *container)
 	return TRUE;
 }
 
+/**
+ * gsf_output_set_error :
+ * @output :
+ * @container :
+ *
+ * <protected> This is a utility routine that should only be used by derived
+ * outputs.
+ *
+ * Returns : Always returns FALSE to facilitate its use.
+ */
+gboolean
+gsf_output_set_error (GsfOutput  *output,
+		      gint        code,
+		      char const *format,
+		      ...)
+{
+	g_return_val_if_fail (GSF_IS_OUTPUT (output), FALSE);
+
+	g_clear_error (&output->err);
+
+	if (format != NULL) {
+		va_list args;
+		va_start (args, format);
+		output->err = g_new (GError, 1);
+		output->err->domain = gsf_output_error_id ();
+		output->err->code = code;
+		output->err->message = g_strdup_vprintf (format, args);
+		va_end (args);
+	}
+
+	return FALSE;
+}
+
 static void
 cb_output_wrap_screwup (gpointer wrapee, GObject *wrapper)
 {
@@ -320,10 +393,60 @@ gsf_output_unwrap (GsfOutput *wrapper, GsfOutput *wrapee)
 }
 
 GQuark
-gsf_output_error (void)
+gsf_output_error_id (void)
 {
 	static GQuark quark;
 	if (!quark)
 		quark = g_quark_from_static_string ("gsf_output_error");
 	return quark;
 }
+
+static gboolean
+gsf_output_vprintf (GsfOutput *output, char const *fmt, va_list args)
+{
+	int reslen;
+
+	if (NULL == output->printf_buf) {
+		output->printf_buf_size = 128;
+		output->printf_buf = g_malloc (output->printf_buf_size);
+	}
+	reslen = g_vsnprintf (output->printf_buf, output->printf_buf_size, fmt, args);
+
+	/* handle C99 or older -1 case of vsnprintf */
+	if (reslen < 0 || reslen >= output->printf_buf_size) {
+		g_free (output->printf_buf);
+		output->printf_buf = g_strdup_vprintf (fmt, args);
+		reslen = output->printf_buf_size = strlen (output->printf_buf);
+	}
+
+	return gsf_output_write (output, reslen, output->printf_buf);
+}
+
+gboolean
+gsf_output_printf (GsfOutput *output, char const* format, ...)
+{
+	va_list args;
+	gboolean ret;
+
+	g_return_val_if_fail (output != NULL, FALSE);
+	g_return_val_if_fail (format != NULL, FALSE);
+	g_return_val_if_fail (strlen (format) > 0, FALSE);
+
+	va_start (args, format);
+	ret = GET_CLASS (output)->Vprintf (output, format, args);
+	va_end (args);
+
+	return ret;
+}
+
+/* Like fputs, this assumes that the line already ends with a newline */
+gboolean
+gsf_output_puts (GsfOutput *output, char const* line)
+{
+	size_t nbytes = 0;
+
+	g_return_val_if_fail (line != NULL, FALSE);
+
+	nbytes = strlen (line);
+	return gsf_output_write (output, nbytes, line);
+}
